/**************************************************************************************************
 *
 * Classname:	Board
 * Properties:	Contains the actual battlefield plus methods to manipulate it during gameplay
 * Superclass:	Object
 * Author:		Peter Asplund & Henrik Hedman 
 *				
 **************************************************************************************************/

import javax.swing.*;

public class Board
{
	private Zone[][] maze;		//A matrix of Zone:s representing the maze
	private ImageIcon[] pics;	//This is an array containing the images for the grapical representation of the maze
	private int walls;
	private int xSize;
	private int ySize;
	private static int startPosX;
	private static int startPosY;
	private static int goalX;
	private static int goalY;
	private List closed;
	private List opened;
	private List temp;
			  
/* 	
 * Function:		Board (Constructor)
 * Purpose:		declare important parameters
 * Arguments:	board size (in x and y plane)
 *	 				the number of walls inside maze
 * Returns:		-
 */	
	public Board(int x_size, int y_size, int number_of_walls)
	{
		xSize = x_size;
		ySize = y_size;
		maze = new Zone[x_size][y_size];
		pics = new ImageIcon[6];
		closed = new List();
		opened = new List();
		temp = new List();
		walls = number_of_walls;
	}
	
/* 	
 * Function:		init_board
 * Purpose:		initiate the maze
 * Arguments:	-
 * Returns:		-
 */	
	public void init_board()
	{
		Misc.load_pics(pics);					//Load the images into the 'pics'-array
		Walllayer.init_array(maze, pics);		//Initiate the maze array
		Walllayer.wallplacer(walls, maze);		//Randomly place the mines on the maze
		Walllayer.markerplacer(maze);			//Places the marker at a random position
	}
	
/* 	
 * Function:		find_pos
 * Purpose:		Retrieve the path from the closed-list and draw it.
 * Arguments:	-
 * Returns:		-
 */		
	public void find_pos()
	{
		int x=0,y=0;
		int j = get_length();
		if(!closed.isEmpty())
		{	
			for(int i = 0; i<j; i++)
			{
				x = closed.getX();
				y = closed.getY();
				closed.remove();
				maze[x][y].setType(2);
				maze[x][y].setLabel(2);
			}
		}
		maze[startPosX][startPosY].setLabel(3);
		maze[goalX][goalY].setLabel(5);
	}

/* 	
 * Function:		mark_path
 * Purpose:		Draw the shortest path using Zone.getParent()
 * Arguments:	-
 * Returns:		-
 */
	public void mark_path()
	{
		int x = goalX;
		int y = goalY;
		int newX = -1;
		int newY = -1;
		int increase = 0;
			  
		  //set labels for path
		  while (!(startPosX == newX && startPosY == newY)) {
					 increase++;
					 newX = maze[x][y].getParentX();
					 newY = maze[x][y].getParentY();
					 maze[newX][newY].setType(4);
					 maze[newX][newY].setLabel(4);
					 System.out.println(increase + ". Current: " + x + ", " + y);
					 x = newX; 
					 y = newY;
					 System.out.println(increase + ". parent: " + newX + ", " + newY);
					 //System.out.println(increase + ". Start: " + startPosX + ", " + startPosY);
		  }
    
    // set start and goal labels
		maze[startPosX][startPosY].setLabel(3);
		maze[goalX][goalY].setLabel(5);		
	}

/*
 * Function:		setWall
 * Purpose:		place wall elements in the maze
 * Arguments:	coordinates of mouseclick
 * Returns:		-
 */
	public void setWall(int x, int y) {
		if(maze[x][y].getType() == 1) {
			maze[x][y].setType(0);
			maze[x][y].setLabel(0);
		}
		else if (maze[x][y].getType() == 0) {
			maze[x][y].setType(1);
			maze[x][y].setLabel(1);			
		}
		else;
	}
	
/*
 * Function:		get_closed_list
 * Purpose:		gives a printout of the closed list
 * Arguments:	-
 * Returns:		-
 */
	public void get_closed_list()
	{
		closed.print();
	}

/*
 * Function:		setStartPos
 * Purpose:		save the startingposition of the board
 * Arguments:	coordinates x and y
 * Returns:		-
 */	
	public static void setStartPos(int x_pos, int y_pos) {
		startPosX = x_pos;
		startPosY = y_pos;
	}
	
/*
 * Function:		calculateH
 * Purpose:		calculate the heuristic distance to the goal position from each point
 * Arguments:	goal coordinates x and y
 * Returns:		Checkvalue if distances have been calculated.
 */
	public int calculateH(int goalPosX, int goalPosY) {
		if(!(maze[goalPosX][goalPosY].getType() == 1)) {		//Don't calculate if the zone is a wall
			for (int x = 0; x < xSize; x++) {
				for (int y = 0; y < ySize; y++) {
					 //The distance is times ten to speed up calculations (only int needs to be used with good enough accuracy)
					maze[x][y].setH((int)Math.ceil( Math.sqrt( ( (goalPosX-x)*(goalPosX-x) + (goalPosY-y)*(goalPosY-y) )  * 10 )));
				}
			}
			return 0;	//proceed with the A* algorithm
		}
		else
			return 1;	//don't do anything
	}

/*
 * Function:		aStar (A*)
 * Purpose:		the pathfinding algorithm
 * Arguments:	the goal coordinates x and y
 * Returns:		boolean wether a path is found
 */
	public boolean aStar(int goalPosX, int goalPosY) {
		//set goalposition
		goalX = goalPosX;
		goalY = goalPosY;
    
		//insert startpos into open list
		opened.insert(startPosX, startPosY);
		
		  boolean pathFound = false;
		  boolean noPath = false;
		  int min;
		  int oldMin;
		  int currentX = -1;
		  int currentY = -1;
		
		while(!pathFound) {
			
		  //set temp min and old min to impossible values
		  min = 1000;
		  oldMin = 1000;
		  if(startPosX==goalPosX && startPosY == goalPosY) {
					  maze[goalPosX][goalPosY].setParent(startPosX,startPosY);
					 pathFound = true;
		  }
			
			//make a copy of openlist when traversing
			temp = opened.copy(opened.getHead());
			
			//Check the openlist for lowest F-cost
			for (int i = 0; i < opened.length(); i++) {
				min = maze[ temp.getX() ][ temp.getY() ].getF();
				if (min <= oldMin) {
					currentX = temp.getX();
					currentY = temp.getY();
					oldMin = min;
				}
				temp.remove();
			}
	    
		  //if no path is found.
		  if(opened.isEmpty()) {
					 pathFound = true;	
					 noPath = true;
		  }
     
		  opened.remove(currentX, currentY);						//remove x,y from openlist
			
			/* insert value into closed list.
			 *
			 *
			 */
		  if (closed.insert(currentX, currentY)) {					//move this to the closed list
				maze[currentX][currentY].setType(4);
		  }

			//insert into openlist
		  for (int x = -1; x <= 1; x++) {
					 for (int y = -1; y <= 1; y++) {
								if(!(currentX + x < 0 || currentY + y < 0 || currentX + x >= get_x_size() || currentY + y >= get_y_size())) {	//check inside bounds
										  if(maze[currentX + x][currentY + y].getType() == 0) {													//walkable
													 if (opened.find(currentX + x, currentY + y)) {												//if in openlist
																if( x == 0 || y == 0) {
																		  if (maze[currentX + x][currentY + y].getG() > (maze[currentX][currentY].getG() + 4)) {
																					 maze[currentX + x][currentY + y].setG(maze[currentX][currentY].getG() + 4);//save the G and H values
																					 maze[currentX + x][currentY + y].setParent(currentX, currentY);
																		  }
																}
																else if (maze[currentX + x][currentY + y].getG() > (maze[currentX][currentY].getG() + 5)) {
																		  maze[currentX + x][currentY + y].setG(maze[currentX][currentY].getG() + 5);			//save the G and H values
																		  maze[currentX + x][currentY + y].setParent(currentX, currentY);
																}
																else;
													 }
													 else {
																opened.insert(currentX + x, currentY + y);								//add both x and y to openlist
																maze[currentX + x][currentY + y].setParent(currentX, currentY);			//mark parent to surrounding squares
																if( x == 0 || y == 0)
																		  maze[currentX + x][currentY + y].setG(maze[currentX][currentY].getG() + 3);			//save the G and H values
																else
																		  maze[currentX + x][currentY + y].setG(maze[currentX][currentY].getG() + 4);			//save the G and H values
																if(opened.getX() == goalPosX && opened.getY() == goalPosY) {
																		  pathFound = true;
																		  closed.insert(currentX+x, currentY+y);
																		  break; 
																}
													 }
										  }
								}
					 }
		  }
		}
    return noPath;
	}
	
/*
 * Function:		get_length
 * Arguments:	-
 * Returns:		length of closed list
 */
	public int get_length()
	{
		int x = closed.length();
		return x;
	}

/*
 * Function:		get_x_size
 * Arguments:	-
 * Returns:		size of maze in x direction
 */
	public int get_x_size()
	{
		return maze.length;
	}

/*
 * Function:		Board (Constructor)
 * Arguments:	-
 * Returns:		size of maze in y direction
 */
	public int get_y_size()
	{
		return maze[0].length;
	}
	
/*
 * Function:		Board (Constructor)
 * Arguments:	-
 * Returns:		the maze
 */
	public Zone[][] get_maze()
	{
		return maze;
	}

/*
 * Function:		Board (Constructor)
 * Arguments:	-
 * Returns:		the number of walls
 */
	public int get_walls()
	{
		return walls;
	}
	
}
